@@ -8,19 +8,20 @@
 
 import collections
 import datetime
+import functools
 import glob
 import itertools
 import os
 import re
 import sqlite3
 import sys
+import threading
 import zlib
 
-from coverage import env
-from coverage.backward import get_thread_id, iitems, to_bytes, to_string
 from coverage.debug import NoDebugging, SimpleReprMixin, clipped_repr
+from coverage.exceptions import CoverageException
 from coverage.files import PathAliases
-from coverage.misc import CoverageException, contract, file_be_gone, filename_suffix, isolate_module
+from coverage.misc import contract, file_be_gone, filename_suffix, isolate_module
 from coverage.numbits import numbits_to_nums, numbits_union, nums_to_numbits
 from coverage.version import __version__
 
@@ -179,6 +180,10 @@ class CoverageData(SimpleReprMixin):
     Data in a :class:`CoverageData` can be serialized and deserialized with
     :meth:`dumps` and :meth:`loads`.
 
+    The methods used during the coverage.py collection phase
+    (:meth:`add_lines`, :meth:`add_arcs`, :meth:`set_context`, and
+    :meth:`add_file_tracers`) are thread-safe.  Other methods may not be.
+
     """
 
     def __init__(self, basename=None, suffix=None, no_disk=False, warn=None, debug=None):
@@ -207,6 +212,8 @@ class CoverageData(SimpleReprMixin):
         # Maps thread ids to SqliteDb objects.
         self._dbs = {}
         self._pid = os.getpid()
+        # Synchronize the operations used during collection.
+        self._lock = threading.Lock()
 
         # Are we in sync with the data file?
         self._have_used = False
@@ -218,6 +225,15 @@ class CoverageData(SimpleReprMixin):
         self._current_context_id = None
         self._query_context_ids = None
 
+    def _locked(method):            # pylint: disable=no-self-argument
+        """A decorator for methods that should hold self._lock."""
+        @functools.wraps(method)
+        def _wrapped(self, *args, **kwargs):
+            with self._lock:
+                # pylint: disable=not-callable
+                return method(self, *args, **kwargs)
+        return _wrapped
+
     def _choose_filename(self):
         """Set self._filename based on inited attributes."""
         if self._no_disk:
@@ -244,8 +260,8 @@ class CoverageData(SimpleReprMixin):
         Initializes the schema and certain metadata.
         """
         if self._debug.should('dataio'):
-            self._debug.write("Creating data file {!r}".format(self._filename))
-        self._dbs[get_thread_id()] = db = SqliteDb(self._filename, self._debug)
+            self._debug.write(f"Creating data file {self._filename!r}")
+        self._dbs[threading.get_ident()] = db = SqliteDb(self._filename, self._debug)
         with db:
             db.executescript(SCHEMA)
             db.execute("insert into coverage_schema (version) values (?)", (SCHEMA_VERSION,))
@@ -261,13 +277,13 @@ class CoverageData(SimpleReprMixin):
     def _open_db(self):
         """Open an existing db file, and read its metadata."""
         if self._debug.should('dataio'):
-            self._debug.write("Opening data file {!r}".format(self._filename))
-        self._dbs[get_thread_id()] = SqliteDb(self._filename, self._debug)
+            self._debug.write(f"Opening data file {self._filename!r}")
+        self._dbs[threading.get_ident()] = SqliteDb(self._filename, self._debug)
         self._read_db()
 
     def _read_db(self):
         """Read the metadata from a database so that we are ready to use it."""
-        with self._dbs[get_thread_id()] as db:
+        with self._dbs[threading.get_ident()] as db:
             try:
                 schema_version, = db.execute_one("select version from coverage_schema")
             except Exception as exc:
@@ -275,7 +291,7 @@ class CoverageData(SimpleReprMixin):
                     "Data file {!r} doesn't seem to be a coverage data file: {}".format(
                         self._filename, exc
                     )
-                )
+                ) from exc
             else:
                 if schema_version != SCHEMA_VERSION:
                     raise CoverageException(
@@ -293,15 +309,15 @@ class CoverageData(SimpleReprMixin):
 
     def _connect(self):
         """Get the SqliteDb object to use."""
-        if get_thread_id() not in self._dbs:
+        if threading.get_ident() not in self._dbs:
             if os.path.exists(self._filename):
                 self._open_db()
             else:
                 self._create_db()
-        return self._dbs[get_thread_id()]
+        return self._dbs[threading.get_ident()]
 
     def __nonzero__(self):
-        if (get_thread_id() not in self._dbs and not os.path.exists(self._filename)):
+        if (threading.get_ident() not in self._dbs and not os.path.exists(self._filename)):
             return False
         try:
             with self._connect() as con:
@@ -320,6 +336,10 @@ class CoverageData(SimpleReprMixin):
         suitable for use with :meth:`loads` in the same version of
         coverage.py.
 
+        Note that this serialization is not what gets stored in coverage data
+        files.  This method is meant to produce bytes that can be transmitted
+        elsewhere and then deserialized with :meth:`loads`.
+
         Returns:
             A byte string of serialized data.
 
@@ -327,17 +347,20 @@ class CoverageData(SimpleReprMixin):
 
         """
         if self._debug.should('dataio'):
-            self._debug.write("Dumping data from data file {!r}".format(self._filename))
+            self._debug.write(f"Dumping data from data file {self._filename!r}")
         with self._connect() as con:
-            return b'z' + zlib.compress(to_bytes(con.dump()))
+            return b'z' + zlib.compress(con.dump().encode("utf8"))
 
     @contract(data='bytes')
     def loads(self, data):
-        """Deserialize data from :meth:`dumps`
+        """Deserialize data from :meth:`dumps`.
 
         Use with a newly-created empty :class:`CoverageData` object.  It's
         undefined what happens if the object already has data in it.
 
+        Note that this is not for reading data from a coverage data file.  It
+        is only for use on data you produced with :meth:`dumps`.
+
         Arguments:
             data: A byte string of serialized data produced by :meth:`dumps`.
 
@@ -345,13 +368,13 @@ class CoverageData(SimpleReprMixin):
 
         """
         if self._debug.should('dataio'):
-            self._debug.write("Loading data into data file {!r}".format(self._filename))
+            self._debug.write(f"Loading data into data file {self._filename!r}")
         if data[:1] != b'z':
             raise CoverageException(
-                "Unrecognized serialization: {!r} (head of {} bytes)".format(data[:40], len(data))
+                f"Unrecognized serialization: {data[:40]!r} (head of {len(data)} bytes)"
                 )
-        script = to_string(zlib.decompress(data[1:]))
-        self._dbs[get_thread_id()] = db = SqliteDb(self._filename, self._debug)
+        script = zlib.decompress(data[1:]).decode("utf8")
+        self._dbs[threading.get_ident()] = db = SqliteDb(self._filename, self._debug)
         with db:
             db.executescript(script)
         self._read_db()
@@ -381,6 +404,7 @@ class CoverageData(SimpleReprMixin):
             else:
                 return None
 
+    @_locked
     def set_context(self, context):
         """Set the current context for future :meth:`add_lines` etc.
 
@@ -391,7 +415,7 @@ class CoverageData(SimpleReprMixin):
 
         """
         if self._debug.should('dataop'):
-            self._debug.write("Setting context: %r" % (context,))
+            self._debug.write(f"Setting context: {context!r}")
         self._current_context = context
         self._current_context_id = None
 
@@ -422,12 +446,13 @@ class CoverageData(SimpleReprMixin):
         """
         return self._filename
 
+    @_locked
     def add_lines(self, line_data):
         """Add measured line data.
 
-        `line_data` is a dictionary mapping file names to dictionaries::
+        `line_data` is a dictionary mapping file names to iterables of ints::
 
-            { filename: { lineno: None, ... }, ...}
+            { filename: { line1, line2, ... }, ...}
 
         """
         if self._debug.should('dataop'):
@@ -440,7 +465,7 @@ class CoverageData(SimpleReprMixin):
             return
         with self._connect() as con:
             self._set_context_id()
-            for filename, linenos in iitems(line_data):
+            for filename, linenos in line_data.items():
                 linemap = nums_to_numbits(linenos)
                 file_id = self._file_id(filename, add=True)
                 query = "select numbits from line_bits where file_id = ? and context_id = ?"
@@ -454,12 +479,14 @@ class CoverageData(SimpleReprMixin):
                     (file_id, self._current_context_id, linemap),
                 )
 
+    @_locked
     def add_arcs(self, arc_data):
         """Add measured arc data.
 
-        `arc_data` is a dictionary mapping file names to dictionaries::
+        `arc_data` is a dictionary mapping file names to iterables of pairs of
+        ints::
 
-            { filename: { (l1,l2): None, ... }, ...}
+            { filename: { (l1,l2), (l1,l2), ... }, ...}
 
         """
         if self._debug.should('dataop'):
@@ -472,7 +499,7 @@ class CoverageData(SimpleReprMixin):
             return
         with self._connect() as con:
             self._set_context_id()
-            for filename, arcs in iitems(arc_data):
+            for filename, arcs in arc_data.items():
                 file_id = self._file_id(filename, add=True)
                 data = [(file_id, self._current_context_id, fromno, tono) for fromno, tono in arcs]
                 con.executemany(
@@ -486,9 +513,9 @@ class CoverageData(SimpleReprMixin):
         assert lines or arcs
         assert not (lines and arcs)
         if lines and self._has_arcs:
-            raise CoverageException("Can't add lines to existing arc data")
+            raise CoverageException("Can't add line measurements to existing branch data")
         if arcs and self._has_lines:
-            raise CoverageException("Can't add arcs to existing line data")
+            raise CoverageException("Can't add branch measurements to existing line data")
         if not self._has_arcs and not self._has_lines:
             self._has_lines = lines
             self._has_arcs = arcs
@@ -498,6 +525,7 @@ class CoverageData(SimpleReprMixin):
                     ('has_arcs', str(int(arcs)))
                 )
 
+    @_locked
     def add_file_tracers(self, file_tracers):
         """Add per-file plugin information.
 
@@ -510,18 +538,18 @@ class CoverageData(SimpleReprMixin):
             return
         self._start_using()
         with self._connect() as con:
-            for filename, plugin_name in iitems(file_tracers):
+            for filename, plugin_name in file_tracers.items():
                 file_id = self._file_id(filename)
                 if file_id is None:
                     raise CoverageException(
-                        "Can't add file tracer data for unmeasured file '%s'" % (filename,)
+                        f"Can't add file tracer data for unmeasured file '{filename}'"
                     )
 
                 existing_plugin = self.file_tracer(filename)
                 if existing_plugin:
                     if existing_plugin != plugin_name:
                         raise CoverageException(
-                            "Conflicting file tracer name for '%s': %r vs %r" % (
+                            "Conflicting file tracer name for '{}': {!r} vs {!r}".format(
                                 filename, existing_plugin, plugin_name,
                             )
                         )
@@ -546,7 +574,7 @@ class CoverageData(SimpleReprMixin):
         to associate the right filereporter, etc.
         """
         if self._debug.should('dataop'):
-            self._debug.write("Touching %r" % (filenames,))
+            self._debug.write(f"Touching {filenames!r}")
         self._start_using()
         with self._connect(): # Use this to get one transaction.
             if not self._has_arcs and not self._has_lines:
@@ -565,7 +593,7 @@ class CoverageData(SimpleReprMixin):
         re-map paths to match the local machine's.
         """
         if self._debug.should('dataop'):
-            self._debug.write("Updating with data from %r" % (
+            self._debug.write("Updating with data from {!r}".format(
                 getattr(other_data, '_filename', '???'),
             ))
         if self._has_lines and other_data._has_arcs:
@@ -668,7 +696,7 @@ class CoverageData(SimpleReprMixin):
                 # If there is no tracer, there is always the None tracer.
                 if this_tracer is not None and this_tracer != other_tracer:
                     raise CoverageException(
-                        "Conflicting file tracer name for '%s': %r vs %r" % (
+                        "Conflicting file tracer name for '{}': {!r} vs {!r}".format(
                             path, this_tracer, other_tracer
                         )
                     )
@@ -737,7 +765,7 @@ class CoverageData(SimpleReprMixin):
         if self._no_disk:
             return
         if self._debug.should('dataio'):
-            self._debug.write("Erasing data file {!r}".format(self._filename))
+            self._debug.write(f"Erasing data file {self._filename!r}")
         file_be_gone(self._filename)
         if parallel:
             data_dir, local = os.path.split(self._filename)
@@ -745,7 +773,7 @@ class CoverageData(SimpleReprMixin):
             pattern = os.path.join(os.path.abspath(data_dir), localdot)
             for filename in glob.glob(pattern):
                 if self._debug.should('dataio'):
-                    self._debug.write("Erasing parallel data file {!r}".format(filename))
+                    self._debug.write(f"Erasing parallel data file {filename!r}")
                 file_be_gone(filename)
 
     def read(self):
@@ -784,7 +812,7 @@ class CoverageData(SimpleReprMixin):
         """
         self._start_using()
         with self._connect() as con:
-            contexts = set(row[0] for row in con.execute("select distinct(context) from context"))
+            contexts = {row[0] for row in con.execute("select distinct(context) from context")}
         return contexts
 
     def file_tracer(self, filename):
@@ -843,7 +871,7 @@ class CoverageData(SimpleReprMixin):
             self._query_context_ids = None
 
     def lines(self, filename):
-        """Get the list of lines executed for a file.
+        """Get the list of lines executed for a source file.
 
         If the file was not measured, returns None.  A file might be measured,
         and have no lines executed, in which case an empty list is returned.
@@ -857,7 +885,7 @@ class CoverageData(SimpleReprMixin):
             arcs = self.arcs(filename)
             if arcs is not None:
                 all_lines = itertools.chain.from_iterable(arcs)
-                return list(set(l for l in all_lines if l > 0))
+                return list({l for l in all_lines if l > 0})
 
         with self._connect() as con:
             file_id = self._file_id(filename)
@@ -964,13 +992,16 @@ class CoverageData(SimpleReprMixin):
         """
         with SqliteDb(":memory:", debug=NoDebugging()) as db:
             temp_store = [row[0] for row in db.execute("pragma temp_store")]
-            compile_options = [row[0] for row in db.execute("pragma compile_options")]
+            copts = [row[0] for row in db.execute("pragma compile_options")]
+            # Yes, this is overkill. I don't like the long list of options
+            # at the end of "debug sys", but I don't want to omit information.
+            copts = ["; ".join(copts[i:i + 3]) for i in range(0, len(copts), 3)]
 
         return [
             ('sqlite3_version', sqlite3.version),
             ('sqlite3_sqlite_version', sqlite3.sqlite_version),
             ('sqlite3_temp_store', temp_store),
-            ('sqlite3_compile_options', compile_options),
+            ('sqlite3_compile_options', copts),
         ]
 
 
@@ -995,28 +1026,14 @@ class SqliteDb(SimpleReprMixin):
         if self.con is not None:
             return
 
-        # SQLite on Windows on py2 won't open a file if the filename argument
-        # has non-ascii characters in it.  Opening a relative file name avoids
-        # a problem if the current directory has non-ascii.
-        filename = self.filename
-        if env.WINDOWS and env.PY2:
-            try:
-                filename = os.path.relpath(self.filename)
-            except ValueError:
-                # ValueError can be raised under Windows when os.getcwd() returns a
-                # folder from a different drive than the drive of self.filename in
-                # which case we keep the original value of self.filename unchanged,
-                # hoping that we won't face the non-ascii directory problem.
-                pass
-
         # It can happen that Python switches threads while the tracer writes
         # data. The second thread will also try to write to the data,
         # effectively causing a nested context. However, given the idempotent
         # nature of the tracer operations, sharing a connection among threads
         # is not a problem.
         if self.debug:
-            self.debug.write("Connecting to {!r}".format(self.filename))
-        self.con = sqlite3.connect(filename, check_same_thread=False)
+            self.debug.write(f"Connecting to {self.filename!r}")
+        self.con = sqlite3.connect(self.filename, check_same_thread=False)
         self.con.create_function('REGEXP', 2, _regexp)
 
         # This pragma makes writing faster. It disables rollbacks, but we never need them.
@@ -1047,14 +1064,14 @@ class SqliteDb(SimpleReprMixin):
                 self.close()
             except Exception as exc:
                 if self.debug:
-                    self.debug.write("EXCEPTION from __exit__: {}".format(exc))
+                    self.debug.write(f"EXCEPTION from __exit__: {exc}")
                 raise
 
     def execute(self, sql, parameters=()):
         """Same as :meth:`python:sqlite3.Connection.execute`."""
         if self.debug:
-            tail = " with {!r}".format(parameters) if parameters else ""
-            self.debug.write("Executing {!r}{}".format(sql, tail))
+            tail = f" with {parameters!r}" if parameters else ""
+            self.debug.write(f"Executing {sql!r}{tail}")
         try:
             try:
                 return self.con.execute(sql, parameters)
@@ -1078,8 +1095,8 @@ class SqliteDb(SimpleReprMixin):
             except Exception:
                 pass
             if self.debug:
-                self.debug.write("EXCEPTION from execute: {}".format(msg))
-            raise CoverageException("Couldn't use data file {!r}: {}".format(self.filename, msg))
+                self.debug.write(f"EXCEPTION from execute: {msg}")
+            raise CoverageException(f"Couldn't use data file {self.filename!r}: {msg}") from exc
 
     def execute_one(self, sql, parameters=()):
         """Execute a statement and return the one row that results.
@@ -1096,14 +1113,20 @@ class SqliteDb(SimpleReprMixin):
         elif len(rows) == 1:
             return rows[0]
         else:
-            raise CoverageException("Sql {!r} shouldn't return {} rows".format(sql, len(rows)))
+            raise CoverageException(f"Sql {sql!r} shouldn't return {len(rows)} rows")
 
     def executemany(self, sql, data):
         """Same as :meth:`python:sqlite3.Connection.executemany`."""
         if self.debug:
             data = list(data)
-            self.debug.write("Executing many {!r} with {} rows".format(sql, len(data)))
-        return self.con.executemany(sql, data)
+            self.debug.write(f"Executing many {sql!r} with {len(data)} rows")
+        try:
+            return self.con.executemany(sql, data)
+        except Exception:
+            # In some cases, an error might happen that isn't really an
+            # error.  Try again immediately.
+            # https://github.com/nedbat/coveragepy/issues/1010
+            return self.con.executemany(sql, data)
 
     def executescript(self, script):
         """Same as :meth:`python:sqlite3.Connection.executescript`."""
